function [x,flag,relres,iter,relvec] = Gmres(A,b,restart,tol,maxit,x0)
% 
% Ԥ GMRES 
% 
% Inputs:
%   A -- coefficient matrix or function for matrix-vector production
%   b -- right hand side 
%   restart -- step number for restarting 
%   tol -- stopping criterion 
%        (optional, if tol is [] or not given, default is 1e-8)
%   maxit -- maximum iteration 
%        (optional, if maxit is [] or not given, default is length(b))
%   x0 -- initial guess 
%        (optional, if x is [] or not given, default is zero vector)
%
% Outputs:
%   x -- approximate solution
%   flag -- convergence flag
%   relres -- relative residual
%   iter -- iteration number
%   relvec -- vector containing relative residuals of each iterations
%

if nargin < 2, help(mfilename), error(mfilename), end
if nargin < 6
    x0 = zeros(length(b),1); 
end
if nargin < 5 || isempty(maxit)
    maxit = length(b);
end
if nargin < 4 || isempty(tol)
    tol = 1e-8;
end
if nargin < 3 || isempty(restart)
    restart = length(b);
end

% initialization
flag = 0;
m = restart;  
n = length(b);
iter  = 0;    
relres = 0.0;  
x = x0;

% initialize workspace
xi = zeros(n,1);  xi(1) = 1.0;
cs = zeros(m,1);  sn = zeros(m,1);

% initialize workspace for restarted GMRES
if (m < maxit) 
    H(1:m+1,1:m) = zeros(m+1,m);
    V(1:n,1:m+1) = zeros(n,m+1);
end

bnrm2 = norm(b);
if  (bnrm2 == 0.0), bnrm2 = 1.0; end

% Quick check of approximation
r = b - A*x;  % r = b - Ax(x);
res = norm(r);  % residual 
relres = res / bnrm2;  % relative residual
relvec = relres;
if (relres < tol),  return, end
% fprintf(' iter=%d, relres=%e\n', iter, norm(r));

% begin iteration
for i_outer = 1 : ceil(maxit/restart)
    V(:,1) = r / res;        % V(:,1) = r / norm(r);
    s = res * xi;            % s = norm(r)*xi;
    for i = 1 : m            % construct orthonormal bases using MGS
        w = A*V(:,i);           % w = A *V(:,i)
        for k = 1 : i 
            H(k,i)= w'*V(:,k);
            w = w - H(k,i)*V(:,k);
        end
        H(i+1,i) = norm(w);
        V(:,i+1) = w / H(i+1,i);
        for k = 1 : i-1 % Apply Givens rotation
            temp     =   cs(k)*H(k,i) + sn(k)*H(k+1,i);
            H(k+1,i) = - sn(k)*H(k,i) + cs(k)*H(k+1,i);
            H(k,i)   =   temp;
        end
        % form i-th rotation matrix
        [cs(i), sn(i)] = rotmat( H(i,i), H(i+1,i) );
        % approximate residual norm
        temp   =   cs(i)*s(i);
        s(i+1) = - sn(i)*s(i);
        s(i)   =   temp;
        H(i,i) =   cs(i)*H(i,i) + sn(i)*H(i+1,i);
        H(i+1,i) = 0.0;
        relres  = abs(s(i+1)) / bnrm2;  % check convergence
        relvec = [relvec; relres];
        % fprintf(' iter=%d, relres=%e\n', i, relres);
        if ( relres < tol )
            y = H(1:i,1:i) \ s(1:i);
            z = V(:,1:i) * y; 
            x = x + z; 
            break;
        end
    end
    if ( relres < tol ), break, end
    y = H(1:m,1:m) \ s(1:m);
    z = V(:,1:m)*y;  
    x = x + z;  
    r = b - A*x; % r = b - Ax(x)
    res = norm(r);  % check convergence
    relres = res / bnrm2;
    % fprintf(' iter=%d, relres=%e\n', i, relres);
    if ( relres < tol ), break, end;
end

iter = (i_outer - 1) * m + i;  

if ( relres > tol ), flag=1; end  % no convergence 

%%%%% rotmat ==========================================================
function [ c, s ] = rotmat( a, b )
if ( b == 0.0 )
   c = 1.0;
   s = 0.0;
elseif ( abs(b) > abs(a) )
   tau = a / b;
   s = 1.0 / sqrt( 1.0 + tau^2 );
   c = tau * s;
else
   tau = b / a;
   c = 1.0 / sqrt( 1.0 + tau^2 );
   s = tau * c;
end
%%%%% END =============================================================



